local lg, lm = love.graphics, love.mouse
local sqrt, min, max = math.sqrt, math.min, math.max

-- AABB --

local mtAABB = {}

function mtAABB:center()
    return self.x + self.w * 0.5, self.y + self.h * 0.5
end

function mtAABB:bbox(rb)
    if rb then
        return self.x, self.y, self.x + self.w, self.y + self.h
    else
        return self.x, self.y, self.w, self.h        
    end
end

function mtAABB:move(dx, dy)
    self.x, self.y = self.x + dx, self.y + dy
end

function mtAABB:contains(x, y)
    local l, t, r, b = self:bbox(true)
    return x >= l and x <= r and y >= t and y <= b
end

function mtAABB:draw(dx, dy)
    dx, dy = dx or 0, dy or 0
    lg.rectangle('line', self.x + dx, self.y + dy, self.w, self.h)
end

-- new rectangle
local function AABB(x, y, w, h)
    return setmetatable({x = x, y = y, w = w, h = h}, {__index = mtAABB})
end

-- COLLIDE

-- thanks ivan
local function segmentVsRect(x1, y1, x2, y2, l, t, r, b)     
    local dx, dy = x2 - x1, y2 - y1
    local d = sqrt(dx * dx + dy * dy)
    if d == 0 then return false end
    local nx, ny = dx / d, dy / d
        
    local tmin, tmax = 0, d
        
    if nx == 0 then
        if x1 < l or x1 > r then return end
    else
        local t1, t2 = (l - x1) / nx, (r - x1) / nx
        if t1 > t2 then t1, t2 = t2, t1 end
        tmin = max(tmin, t1)
        tmax = min(tmax, t2)
        if tmin > tmax then return end
    end
        
    if ny == 0 then
        if y1 < t or y1 > b then return end
    else
        local t1, t2 = (t - y1) / ny, (b - y1) / ny
        if t1 > t2 then t1, t2 = t2, t1 end
        tmin = max(tmin, t1)
        tmax = min(tmax, t2)
        if tmin > tmax then return end
    end
        
    local qx, qy = x1 + nx * tmin, y1 + ny * tmin
    nx, ny = qx == l and -1 or (qx == r and 1 or 0), qy == t and -1 or (qy == b and 1 or 0)
    
    return true, qx, qy, nx, ny
end

local function rectXrect(l1, t1, w1, h1, l2, t2, w2, h2, gx, gy)
    gx, gy = gx + 0.5 * w1, gy + 0.5* h1
    
    local cx, cy = l1 + 0.5 * w1, t1 + 0.5 * h1 -- center of first rectangle
    local ox, oy = l2 + 0.5 * w2, t2 + 0.5 * h2 -- center of second rectangle
    local hw, hh = (w1 + w2) * 0.5, (h1 + h2) * 0.5
    
    --[[
    lg.setColor(0, 1, 1, 0.2)
    lg.rectangle('line', ox - hw, oy - hh, hw * 2, hh * 2)
    ]]
    
    -- check intersection of line (cx, cy, gx, gy) with rect - center in (ox;oy), width = w1 + w2, height = h1 + h2
    local overlaps, px, py, nx, ny = segmentVsRect(cx, cy, gx, gy, ox - hw, oy - hh, ox + hw, oy + hh)
    
    if overlaps then
        px = px - 0.5 * w1 -- + ny
        py = py - 0.5 * h1 -- + ny
        return {
            px = px, py = py, --new position
            nx = nx, ny = ny, -- normal
            dx = px - l1, dy = py - t1, -- correction move
            --d = sq_dist(px, py, cx, cy)
        }
    end
end

local function projectAABB(item, items, goalX, goalY)
    local l, t, w, h = item:bbox()
    
    local l2, t2, w2, h2, col
    for _, other in pairs(items) do
        if other ~= item then
            l2, t2, w2, h2 = other:bbox()
            col = rectXrect(l, t, w, h, l2, t2, w2, h2, goalX, goalY)
            if col then
                lg.setColor(0, 1, 0)
                item:draw(col.dx, col.dy)
            end
        end
    end
end


-- LET'S MAKE LOVE

local items, selected

function love.load()
    items = {}
    
    items[1] = AABB(100, 100, 50, 50)
    items[2] = AABB(400, 400, 60, 100)
end

function love.draw()
    lg.setColor(1, 1, 1)    
    for i = 1, #items do
        items[i]:draw()
    end
    
    if selected and lm.isDown(2) then
        local goalX, goalY = lm.getPosition()
        lg.setColor(1, 0, 1)
        lg.line(selected.x, selected.y, goalX, goalY)
        selected:draw(goalX - selected.x, goalY - selected.y)
        projectAABB(selected, items, goalX, goalY)
    end
    
    lg.setColor(1, 1, 1)
    lg.print('mouse button 1 - select and move item\nmouse button 2 - select and project item')
end

function love.mousepressed(x, y, button)
    if button == 1 or button == 2 then
        selected = nil
        for i = 1, #items do
            if items[i]:contains(x, y) then
                selected = items[i]
                break
            end
        end
    end
end

function love.mousereleased(x, y, button)
    if (button == 1 or button == 2) and selected then
        selected = nil
    end
end

function love.mousemoved(x, y, dx, dy)
    if lm.isDown(1) and selected then
        selected:move(dx, dy)
    end
end

function love.keypressed(key)
    if key == 'escape' then
        love.event.quit()
    end    
end